@@ -21,6 +21,7 @@
 
 #include <stdlib.h>
 #include <string.h>
+#include <ctype.h>
 #include <math.h>
 #include <assert.h>
 
@@ -132,6 +133,12 @@ mp_set_from_float(float rx, int *z)
     return;
 }
 
+void
+mp_set_from_random(int t[MP_SIZE])
+{
+    mp_set_from_double(drand48(), t);
+}
+
 /*  CONVERTS DOUBLE-PRECISION NUMBER DX TO MULTIPLE-PRECISION Z.
  *  SOME NUMBERS WILL NOT CONVERT EXACTLY ON MACHINES
  *  WITH BASE OTHER THAN TWO, FOUR OR SIXTEEN.
@@ -468,39 +475,132 @@ mp_cast_to_double(const int *x)
     }
 }
 
-static int
-char_val(char chr)
+
+/* Convert MP number to fixed number string in the given base to the
+ * maximum number of digits specified.
+ */
+void
+mp_cast_to_string(char *target, int target_len, const int *MPnumber, int base, int accuracy)
 {
-    if (chr >= '0' && chr <= '9') {
-        return(chr - '0');
-    } else if (chr >= 'a' && chr <= 'f') {
-        return(chr - 'a' + 10);
-    } else if (chr >= 'A' && chr <= 'F') {
-        return(chr - 'A' + 10);
-    } else {
-        return(-1);
+    static char digits[] = "0123456789ABCDEF";
+    char *optr, *start, *end, *last_non_zero;
+    int number[MP_SIZE], integer_component[MP_SIZE], fractional_component[MP_SIZE], MPbase[MP_SIZE], temp[MP_SIZE];
+   
+    optr = target;
+
+    /* Insert sign */
+    if (mp_is_negative(MPnumber)) {
+        *optr++ = '-';
+        mp_abs(MPnumber, number);
+    } else  {
+        mp_set_from_mp(MPnumber, number);	
     }
+   
+    /* Add rounding factor */
+    mp_set_from_integer(base, MPbase);
+    mppwr(MPbase, -(accuracy+1), temp);
+    mpmuli(temp, base, temp);
+    mpdivi(temp, 2, temp);
+    mp_add(number, temp, number);
+
+    /* Split into integer and fractional component */
+    mpcmim(number, integer_component);
+    mpcmf(number, fractional_component);  
+
+    /* Write out the integer component least significant digit to most */
+    start = optr;
+    mp_set_from_mp(integer_component, temp);
+    do {
+        int t[MP_SIZE], t2[MP_SIZE], t3[MP_SIZE];
+       
+        mpdivi(temp, base, t);
+        mpcmim(t, t);
+        mpmuli(t, base, t2);
+       
+        mp_subtract(temp, t2, t3);
+        mpcmim(t3, t3);
+        *optr++ = digits[mp_cast_to_int(t3)];
+       
+        mp_set_from_mp(t, temp);
+    } while (!mp_is_zero(temp));
+    end = optr - 1;
+   
+    /* Reverse digits */
+    while(start < end) {
+        char t;
+        t = *start;
+        *start = *end;
+        *end = t;
+        start++;
+        end--;
+    }
+   
+    /* Stop if there is no fractional component or not showing fractional part */
+    if ((mp_is_zero(fractional_component) && !v->show_zeroes) || accuracy == 0) {
+        *optr = '\0';
+        return;
+    }
+   
+    last_non_zero = optr;
+    *optr++ = '.';
+   
+    /* Write out the fractional component */
+    mp_set_from_mp(fractional_component, temp);
+    do {
+        int d;
+        int digit[MP_SIZE];
+
+        mpmuli(temp, base, temp);
+        mpcmim(temp, digit);
+        d = mp_cast_to_int(digit);
+       
+        *optr++ = digits[d];
+        if(d != 0)
+	   last_non_zero = optr;
+        mp_subtract(temp, digit, temp);
+        accuracy--;
+    } while (!mp_is_zero(temp) && accuracy > 0);
+
+    /* Strip trailing zeroes */
+    if (!v->show_zeroes)
+       optr = last_non_zero;
+
+    *optr = '\0';
 }
 
+
+static int
+char_val(char chr, int base)
+{
+    int value;
+    if (chr >= '0' && chr <= '9') {
+        value = chr - '0';
+    } else if (chr >= 'a' && chr <= 'f') {
+        value = chr - 'a' + 10;
+    } else if (chr >= 'A' && chr <= 'F') {
+        value = chr - 'A' + 10;
+    } else {
+        return -1;
+    }
+    if (value >= base)
+       return -1;
+    return value;
+}
+
+
 /* Convert string into an MP number, in the given base
  */
 void
-MPstr_to_num(const char *str, int base, int *MPval)
+mp_set_from_string(const char *str, int base, int *MPval)
 {
     const char *optr;
-    int MP1[MP_SIZE], MP2[MP_SIZE], MPbase[MP_SIZE];
-    int i, inum;
-    int exp      = 0;
-    int exp_sign = 1;
+    int inum;
     int negate = 0;
 
-    mp_set_from_integer(0, MPval);
-    mp_set_from_integer(base, MPbase);
-
     optr = str;
 
-    /* Remove any initial spaces or tabs. */
-    while (*optr == ' ' || *optr == '\t') {
+    /* Remove leading whitespace */
+    while (isspace(*optr)) {
         optr++;
     }
 
@@ -510,83 +610,73 @@ MPstr_to_num(const char *str, int base, int *MPval)
         optr++;
     }
 
-    while ((inum = char_val(*optr)) >= 0) {
-        mpmul(MPval, MPbase, MPval);
+    /* Convert integer part */
+    mp_set_from_integer(0, MPval);
+    while ((inum = char_val(*optr, base)) >= 0) {
+        mpmuli(MPval, base, MPval);
         mp_add_integer(MPval, inum, MPval);
         optr++;
     }
-
+   
+    /* Convert fractional part */
     if (*optr == '.' || *optr == *v->radix) {
+        int numerator[MP_SIZE], denominator[MP_SIZE];
+       
         optr++;
-        for (i = 1; (inum = char_val(*optr)) >= 0; i++) {
-            mppwr(MPbase, i, MP1);
-            mp_set_from_integer(inum, MP2);
-            mpdiv(MP2, MP1, MP1);
-            mp_add(MPval, MP1, MPval);
+
+        mp_set_from_integer(0, numerator);
+        mp_set_from_integer(1, denominator);
+        while ((inum = char_val(*optr, base)) >= 0) {
+	    mpmuli(denominator, base, denominator);
+	    mpmuli(numerator, base, numerator);
+	    mp_add_integer(numerator, inum, numerator);
             optr++;
         }
+        mpdiv(numerator, denominator, numerator);
+        mp_add(MPval, numerator, MPval);
+    }
+   
+    /* Convert exponential part */
+    if (*optr == 'e' || *optr == 'E') {
+        int negate = 0;
+        int MPbase[MP_SIZE], MPexponent[MP_SIZE], temp[MP_SIZE];
+
+        optr++;       
+
+        /* Get sign */
+        if (*optr == '-') {
+	    negate = 1;
+	    optr++;
+	} else if (*optr == '+') {
+	    optr++;
+	}
+
+        /* Get magnitude */
+        mp_set_from_integer(0, MPexponent);
+        while ((inum = char_val(*optr, base)) >= 0) {
+            mpmuli(MPexponent, base, MPexponent);
+            mp_add_integer(MPexponent, inum, MPexponent);
+            optr++;
+        }
+        if (negate) {
+            mp_invert_sign(MPexponent, MPexponent);
+        }
+
+        mp_set_from_integer(base, MPbase);       
+        mppwr2(MPbase, MPexponent, temp);
+        mpmul(MPval, temp, MPval);
     }
 
-    while (*optr == ' ') {
+    /* Strip trailing whitespace */
+    while (isspace(*optr)) {
         optr++;
     }
- 
+   
     if (*optr != '\0') {
-        if (*optr == '-') {
-            exp_sign = -1;
-        }
- 
-        while ((inum = char_val(*++optr)) >= 0) {
-            exp = exp * basevals[(int) base] + inum;
-        }
+       // FIXME: Error decoding
     }
-    exp *= exp_sign;
-
+ 
     if (negate == 1) {
         mp_invert_sign(MPval, MPval);
     }
 }
-
-static void 
-calc_xtimestenpowx(int s1[MP_SIZE], int s2[MP_SIZE], int t1[MP_SIZE])
-{
-    int MP1[MP_SIZE], MP2[MP_SIZE];
-
-    mp_set_from_integer(10, MP2);
-    mppwr2(MP2, s2, MP1);
-    mpmul(s1, MP1, t1);
-}
-
-void
-mp_set_from_string(const char *number, int base, int t[MP_SIZE])
-{
-    int i;
-    char *a = NULL;
-    char *b = NULL;
-
-    int MP_a[MP_SIZE];
-    int MP_b[MP_SIZE];
-
-    assert(number);
-    a = strdup(number);
-    assert(a);
-
-    for (i = 0; !((a[i] == 'e') || (a[i] == 'E')); i++) {
-        assert(a[i]);
-    }
-
-    a[i] = 0;
-    b = &a[i+2];
-
-    MPstr_to_num(a, base, MP_a);
-    MPstr_to_num(b, base, MP_b);
-    if (a[i+1] == '-') {
-        int MP_c[MP_SIZE];
-        mp_invert_sign(MP_b, MP_c);
-        calc_xtimestenpowx(MP_a, MP_c, t);
-    } else {
-        calc_xtimestenpowx(MP_a, MP_b, t);
-    }
-
-    free(a);
-}
